unit UMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Menus, ExtCtrls,
  StdCtrls, ComCtrls, GL, GLu, Viewports, Math, Maths, TreeText, MiscUtils;

type

  { TLatheVertex }

  TLatheVertex = record
    P, C: TVector;
  end;

  TLatheQuad = array [0..3] of TLatheVertex;

  { TLathePoint }

  TLathe = object
    Shape: T2DShape;
    Quads: array of TLatheQuad;
    Slices: Integer;
    LightPos: TVector;
    AmbientColor, ObjectColor, LightColor: TVector;
    SharpAngle: Double;
    AmbientShade: Double;
    Scale: Double;
    Smooth: Boolean;

    function ClosestPointAt(const Point: TVector2D; MaxDistance: Double=MaxDouble): Integer;
    procedure CalcMesh;
  end;

  { TMain }

  TMain = class(TForm)
    MainMenu1: TMainMenu;
    mFile: TMenuItem;
    mFileNew: TMenuItem;
    mFileOpen: TMenuItem;
    mFileSave: TMenuItem;
    mFileSaveAs: TMenuItem;
    mFileBar2: TMenuItem;
    mFileExit: TMenuItem;
    mHelp: TMenuItem;
    mHelpAbout: TMenuItem;
    Panel1: TPanel;
    Splitter1: TSplitter;
    Panel2: TPanel;
    Panel3: TPanel;
    Splitter2: TSplitter;
    Viewport1: TViewport;
    ViewportManager1: TViewportManager;
    pbLathe: TPaintBox;
    Label1: TLabel;
    tbScale: TTrackBar;
    cbStyle: TComboBox;
    cbSlices: TComboBox;
    Label2: TLabel;
    cbObjectColor: TColorButton;
    cbAmbientColor: TColorButton;
    Label3: TLabel;
    cbLightColor: TColorButton;
    Label4: TLabel;
    Label5: TLabel;
    tbSharpness: TTrackBar;
    Label6: TLabel;
    Panel4: TPanel;
    pbLightPos: TPaintBox;
    mGrid: TMenuItem;
    mGrid4: TMenuItem;
    mGrid6: TMenuItem;
    mGrid8: TMenuItem;
    mGrid16: TMenuItem;
    mGrid32: TMenuItem;
    mGridNone: TMenuItem;
    Label7: TLabel;
    tbAmbientShade: TTrackBar;
    mView: TMenuItem;
    mViewFloorGrid: TMenuItem;
    mViewBackgroundColor: TMenuItem;
    ColorDialog1: TColorDialog;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    procedure pbLathePaint(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mFileExitClick(Sender: TObject);
    procedure Panel1Resize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure pbLatheMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure pbLatheMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbLatheMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Viewport1Render(Sender: TObject);
    procedure tbSharpnessChange(Sender: TObject);
    procedure tbScaleChange(Sender: TObject);
    procedure cbSlicesChange(Sender: TObject);
    procedure cbStyleChange(Sender: TObject);
    procedure cbObjectColorColorChanged(Sender: TObject);
    procedure cbAmbientColorColorChanged(Sender: TObject);
    procedure cbLightColorColorChanged(Sender: TObject);
    procedure pbLightPosPaint(Sender: TObject);
    procedure pbLightPosMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbLightPosMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure mGridNoneClick(Sender: TObject);
    procedure mGrid4Click(Sender: TObject);
    procedure mGrid6Click(Sender: TObject);
    procedure mGrid8Click(Sender: TObject);
    procedure mGrid16Click(Sender: TObject);
    procedure mGrid32Click(Sender: TObject);
    procedure tbAmbientShadeChange(Sender: TObject);
    procedure mViewBackgroundColorClick(Sender: TObject);
    procedure mViewFloorGridClick(Sender: TObject);
    procedure Viewport1RenderBackground(Sender: TObject);
    procedure mFileNewClick(Sender: TObject);
    procedure mFileOpenClick(Sender: TObject);
    procedure mFileSaveAsClick(Sender: TObject);
    procedure mFileSaveClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure mHelpAboutClick(Sender: TObject);
  private
    procedure SetGridSize(AValue: Integer);
  private
    DBuffer: TBitmap;
    FGridSize: Integer;
    Lathe: TLathe;
    LatheBoxPickResult: (lbprEmpty, lbprPoint);
    LatheBoxPickIndex: Integer;
    LatheBoxState: (lbsNormal, lbsMoving);
    LatheBoxMovingOffset: TVector2D;
    LatheBoxMovingIndex: Integer;
    BgColor: TColor;
    FileName: string;
    procedure NewFile;
    procedure EnsureBitmapSize;
    procedure UpdateLatheBox;
    procedure PickLatheBox(const P: TVector2D);
    procedure DrawLathe;
    procedure Render;
    procedure DeleteVertex(Index: Integer);
    procedure SplitVertex(Index: Integer);
    procedure FillUIFromLathe;
    procedure UpdateCaption;
    procedure SaveToFile(AFileName: string);
    property GridSize: Integer read FGridSize write SetGridSize;
  public
  end;

var
  Main: TMain;

implementation

{$R *.lfm}

{ TLathe }

function TLathe.ClosestPointAt(const Point: TVector2D; MaxDistance: Double): Integer;
var
  Distance, ClosestDistance: Double;
  I: Integer;
begin
  Result:=-1;
  if Length(Shape.Vertices)=0 then Exit;
  ClosestDistance:=MaxDistance;
  for I:=0 to High(Shape.Vertices) do begin
    Distance:=Maths.Distance(Point, Shape.Vertices[I]);
    if Distance < ClosestDistance then begin
      ClosestDistance:=Distance;
      Result:=I;
    end;
  end;
end;

procedure TLathe.CalcMesh;
var
  Base, Here, Next, BaseNorm, HereNorm, NextNorm: array of TVector;
  I, J, Q: Integer;
  SliceInc, BaseY: Double;
  N1, N2, N3, N4: TVector;

  procedure CalcHereNext(Deg1, Deg2: Double);
  var
    Mtx: TMatrix;
    I: Integer;
  begin
    Mtx.Rotation(0, 1, 0, DegToRad(Deg1));
    for I:=0 to High(Base) do begin
      Here[I]:=Mtx.Transformed(Base[I]);
      HereNorm[I]:=Mtx.TransformedNormal(BaseNorm[I]).Normalized;
    end;
    Mtx.Rotation(0, 1, 0, DegToRad(Deg2));
    for I:=0 to High(Base) do begin
      Next[I]:=Mtx.Transformed(Base[I]);
      NextNorm[I]:=Mtx.TransformedNormal(BaseNorm[I]).Normalized;
    end;
  end;

  function CalcLight(const N: TVector): TVector;
  var
    D, DP: Double;
  begin
    DP:=Direction(Vector(0, 0, 0), LightPos).Dot(N);
    D:=Clamp(DP);
    Result:=AmbientColor.Scaled(Clamp(DP + AmbientShade)).Added(LightColor.Scaled(D)).Multiplied(ObjectColor);
  end;

begin
  if Length(Shape.Vertices) < 2 then begin
    SetLength(Quads, 0);
    Exit;
  end;
  SetLength(Quads, (Length(Shape.Vertices) - 1)*Slices*2);
  SetLength(Base, Length(Shape.Vertices));
  SetLength(Here, Length(Shape.Vertices));
  SetLength(Next, Length(Shape.Vertices));
  SetLength(BaseNorm, Length(Shape.Vertices));
  SetLength(HereNorm, Length(Shape.Vertices));
  SetLength(NextNorm, Length(Shape.Vertices));
  for I:=0 to High(Base) do Base[I]:=Vector2DToVector(Shape.Vertices[I]).Scaled(1/16*Scale);
  BaseY:=Base[0].Y;
  for I:=1 to High(Base) do BaseY:=Min(BaseY, Base[I].Y);
  for I:=0 to High(Base) do Base[I].Y -= BaseY;
  for I:=0 to High(Base) - 1 do BaseNorm[I]:=TriangleNormal(Base[I], Base[(I + 1) mod Length(Base)], Base[I].Added(Vector(0, 0, 1)));
  BaseNorm[High(Base)]:=BaseNorm[High(Base) - 1];
  SliceInc:=360/Slices;
  for J:=0 to Slices - 1 do begin
    CalcHereNext(J*SliceInc, (J + 1)*SliceInc);
    Q:=(Length(Shape.Vertices) - 1)*J*2;
    for I:=1 to High(Base) do begin
      Quads[Q][0].P:=Here[I];
      Quads[Q][1].P:=Here[I - 1];
      Quads[Q][2].P:=Next[I - 1];
      Quads[Q][3].P:=Next[I];
      Inc(Q);
      Quads[Q][0].P:=Here[I - 1];
      Quads[Q][1].P:=Here[I];
      Quads[Q][2].P:=Next[I];
      Quads[Q][3].P:=Next[I - 1];
      Inc(Q);
    end;
    Q:=(Length(Shape.Vertices) - 1)*J*2;
    if Smooth then begin
      for I:=1 to High(Base) do begin
        N1:=HereNorm[I];
        N2:=HereNorm[I - 1];
        N3:=NextNorm[I - 1];
        N4:=NextNorm[I];
        if (N1.Dot(N2) + 1)*0.5 < SharpAngle then begin
          N1:=N2;
          N4:=N3;
        end;
        Quads[Q][0].C:=CalcLight(N1);
        Quads[Q][1].C:=CalcLight(N2);
        Quads[Q][2].C:=CalcLight(N3);
        Quads[Q][3].C:=CalcLight(N4);
        Inc(Q);
        Quads[Q][0].C:=CalcLight(N2.Scaled(-1));
        Quads[Q][1].C:=CalcLight(N1.Scaled(-1));
        Quads[Q][2].C:=CalcLight(N4.Scaled(-1));
        Quads[Q][3].C:=CalcLight(N3.Scaled(-1));
        Inc(Q);
      end;
    end else begin
      for I:=1 to High(Base) do begin
        Quads[Q][0].C:=CalcLight(HereNorm[I - 1]);
        Quads[Q][1].C:=Quads[Q][0].C;
        Quads[Q][2].C:=Quads[Q][0].C;
        Quads[Q][3].C:=Quads[Q][0].C;
        Inc(Q);
        Quads[Q][0].C:=CalcLight(HereNorm[I - 1].Scaled(-1));
        Quads[Q][1].C:=Quads[Q][0].C;
        Quads[Q][2].C:=Quads[Q][0].C;
        Quads[Q][3].C:=Quads[Q][0].C;
        Inc(Q);
      end;
    end;
  end;
end;

{ TMain }

procedure TMain.pbLathePaint(Sender: TObject);
begin
  DrawLathe;
end;

procedure TMain.FormDestroy(Sender: TObject);
begin
  NewFile;
  FreeAndNil(DBuffer);
end;

procedure TMain.mFileExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMain.Panel1Resize(Sender: TObject);
begin
  EnsureBitmapSize;
end;

procedure TMain.FormCreate(Sender: TObject);
begin
  NewFile;
end;

procedure TMain.pbLatheMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  P: TVector2D;
begin
  P:=Vector2D(X, pbLathe.Height - 1 - Y);
  case LatheBoxState of
    lbsNormal: begin
      PickLatheBox(P);
      case LatheBoxPickResult of
        lbprEmpty: pbLathe.Cursor:=crCross;
        lbprPoint: pbLathe.Cursor:=crSizeAll;
      end;
    end;
    lbsMoving: begin
      Lathe.Shape.Vertices[LatheBoxMovingIndex]:=P.Added(LatheBoxMovingOffset);
      if ssCtrl in Shift then Lathe.Shape.Vertices[LatheBoxMovingIndex].Snap(GridSize);
      Lathe.Shape.Vertices[LatheBoxMovingIndex].X:=Max(0, Lathe.Shape.Vertices[LatheBoxMovingIndex].X);
      UpdateLatheBox;
      Lathe.CalcMesh;
    end;
  end;
end;

procedure TMain.pbLatheMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P: TVector2D;
begin
  P:=Vector2D(X, pbLathe.Height - 1 - Y);
  if Button=mbLeft then begin
    case LatheBoxState of
      lbsNormal: begin
        pbLathe.Cursor:=crNone;
        case LatheBoxPickResult of
          lbprPoint: begin
            if ssShift in Shift then begin
              DeleteVertex(LatheBoxPickIndex);
              Exit;
            end;
            if ssAlt in Shift then begin
              SplitVertex(LatheBoxPickIndex);
              Exit;
            end;
            LatheBoxMovingIndex:=LatheBoxPickIndex;
            LatheBoxMovingOffset:=Lathe.Shape.Vertices[LatheBoxPickIndex].Subbed(P);
          end;
          lbprEmpty: begin
            Lathe.Shape.Add(P);
            LatheBoxMovingIndex:=High(Lathe.Shape.Vertices);
            LatheBoxMovingOffset.Zero;
            Lathe.CalcMesh;
          end;
        end;
        LatheBoxState:=lbsMoving;
        UpdateLatheBox;
      end;
    end;
  end;
end;

procedure TMain.pbLatheMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P: TVector2D;
begin
  P:=Vector2D(X, pbLathe.Height - 1 - Y);
  if Button=mbLeft then begin
    PickLatheBox(P);
    LatheBoxState:=lbsNormal;
    UpdateLatheBox;
  end;
end;

procedure TMain.Viewport1Render(Sender: TObject);
begin
  Render;
end;

procedure TMain.tbSharpnessChange(Sender: TObject);
begin
  Lathe.SharpAngle:=tbSharpness.Position/1000;
  Lathe.CalcMesh;
end;

procedure TMain.tbScaleChange(Sender: TObject);
begin
  Lathe.Scale:=tbScale.Position/1000;
  Lathe.CalcMesh;
end;

procedure TMain.cbSlicesChange(Sender: TObject);
begin
  Lathe.Slices:=StrToInt(Copy(cbSlices.Text, 1, Pos(' ', cbSlices.Text) - 1));
  Lathe.CalcMesh;
end;

procedure TMain.cbStyleChange(Sender: TObject);
begin
  Lathe.Smooth:=cbStyle.ItemIndex=1;
  Lathe.CalcMesh;
end;

procedure TMain.cbObjectColorColorChanged(Sender: TObject);
begin
  Lathe.ObjectColor:=ColorToVector(cbObjectColor.ButtonColor);
  Lathe.CalcMesh;
end;

procedure TMain.cbAmbientColorColorChanged(Sender: TObject);
begin
  Lathe.AmbientColor:=ColorToVector(cbAmbientColor.ButtonColor);
  Lathe.CalcMesh;
end;

procedure TMain.cbLightColorColorChanged(Sender: TObject);
begin
  Lathe.LightColor:=ColorToVector(cbLightColor.ButtonColor);
  Lathe.CalcMesh;
end;

procedure TMain.pbLightPosPaint(Sender: TObject);
var
  V: TVector;
begin
  with pbLightPos.Canvas do begin
    Brush.Color:=clWhite;
    Rectangle(-10, -10, pbLightPos.Width + 10, pbLightPos.Height + 1);
    Brush.Color:=clBlack;
    Rectangle(pbLightPos.Width div 2 - 3, pbLightPos.Height div 2 - 3, pbLightPos.Width div 2 + 4, pbLightPos.Height div 2 + 4);
    Brush.Color:=clYellow;
    V:=Direction(Vector(0, 0, 0), Lathe.LightPos).Scaled(40);
    Rectangle(pbLightPos.Width div 2 + Round(V.X) - 3, pbLightPos.Height div 2 - Round(V.Y) - 3,
              pbLightPos.Width div 2 + Round(V.X) + 4, pbLightPos.Height div 2 - Round(V.Y) + 4);
  end;
end;

procedure TMain.pbLightPosMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button <> mbLeft then Exit;
  pbLightPosMouseMove(Sender, Shift, X, Y);
end;

procedure TMain.pbLightPosMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  V: TVector;
begin
  if not (ssLeft in Shift) then Exit;
  V:=Direction(Vector(pbLightPos.Width div 2, pbLightPos.Height div 2, 0), Vector(X, Y, 0));
  V.y:=-V.y;
  Lathe.LightPos:=V;
  pbLightPos.Invalidate;
  Lathe.CalcMesh;
end;

procedure TMain.mGridNoneClick(Sender: TObject);
begin
  GridSize:=0;
end;

procedure TMain.mGrid4Click(Sender: TObject);
begin
  GridSize:=4;
end;

procedure TMain.mGrid6Click(Sender: TObject);
begin
  GridSize:=6;
end;

procedure TMain.mGrid8Click(Sender: TObject);
begin
  GridSize:=8;
end;

procedure TMain.mGrid16Click(Sender: TObject);
begin
  GridSize:=16;
end;

procedure TMain.mGrid32Click(Sender: TObject);
begin
  GridSize:=32;
end;

procedure TMain.tbAmbientShadeChange(Sender: TObject);
begin
  Lathe.AmbientShade:=tbAmbientShade.Position/1000;
  Lathe.CalcMesh;
end;

procedure TMain.mViewBackgroundColorClick(Sender: TObject);
begin
  ColorDialog1.Color:=BgColor;
  if ColorDialog1.Execute then BgColor:=ColorDialog1.Color;
end;

procedure TMain.mViewFloorGridClick(Sender: TObject);
begin
  Viewport1.DrawGrid:=mViewFloorGrid.Checked;
  if mViewFloorGrid.Checked then
    Viewport1.DrawAxes:=[axX, axZ]
  else
    Viewport1.DrawAxes:=[];
end;

procedure TMain.Viewport1RenderBackground(Sender: TObject);
var
  C: TVector;
begin
  C:=ColorToVector(BgColor);
  glClearColor(C.x, C.y, C.Z, 1);
  glClear(GL_COLOR_BUFFER_BIT);
end;

procedure TMain.mFileNewClick(Sender: TObject);
begin
  if MessageDlg('New File', 'Are you sure? Any modifications you have made will be lost!', mtConfirmation, mbYesNo, 0) <> mrYes then Exit;
  FileName:='';
  NewFile;
  pbLightPos.Invalidate;
  UpdateCaption;
end;

procedure TMain.mFileOpenClick(Sender: TObject);

  procedure OpenFrom(AFileName: string);
  var
    TT: TTreeText;
    I: Integer;
  begin
    TT:=TTreeText.Create(Self);
    try
      TT.LoadFromFile(AFileName);
      SetLength(Lathe.Shape.Vertices, StrToIntDef(TT['points'], 0));
      for I:=0 to High(Lathe.Shape.Vertices) do begin
        Lathe.Shape.Vertices[I].X:=StrToIntDef(TT['p' + IntToStr(I) + 'x'], 0);
        Lathe.Shape.Vertices[I].Y:=StrToIntDef(TT['p' + IntToStr(I) + 'y'], 0);
      end;
      Lathe.Slices:=StrToIntDef(TT['slices'], 8);
      Lathe.LightPos.X:=StringToFloat(TT['light.x'], Lathe.LightPos.x);
      Lathe.LightPos.Y:=StringToFloat(TT['light.y'], Lathe.LightPos.y);
      Lathe.AmbientColor.X:=StringToFloat(TT['ambient.r'], Lathe.AmbientColor.x);
      Lathe.AmbientColor.Y:=StringToFloat(TT['ambient.g'], Lathe.AmbientColor.y);
      Lathe.AmbientColor.Z:=StringToFloat(TT['ambient.b'], Lathe.AmbientColor.z);
      Lathe.ObjectColor.X:=StringToFloat(TT['object.r'], Lathe.ObjectColor.x);
      Lathe.ObjectColor.Y:=StringToFloat(TT['object.g'], Lathe.ObjectColor.y);
      Lathe.ObjectColor.Z:=StringToFloat(TT['object.b'], Lathe.ObjectColor.z);
      Lathe.LightColor.X:=StringToFloat(TT['light.r'], Lathe.LightColor.x);
      Lathe.LightColor.Y:=StringToFloat(TT['light.g'], Lathe.LightColor.y);
      Lathe.LightColor.Z:=StringToFloat(TT['light.b'], Lathe.LightColor.z);
      Lathe.SharpAngle:=StringToFloat(TT['sharp'], Lathe.SharpAngle);
      Lathe.AmbientShade:=StringToFloat(TT['ambient.shade'], Lathe.AmbientShade);
      Lathe.Scale:=StringToFloat(TT['scale'], Lathe.Scale);
      Lathe.Smooth:=StrToBoolDef(TT['smooth'], Lathe.Smooth);
      BgColor:=VectorToColor(Vector(StringToFloat(TT['bgcolor.r']), StringToFloat(TT['bgcolor.g']), StringToFloat(TT['bgcolor.b'])));
      mViewFloorGrid.Checked:=StrToBoolDef(TT['floorgrid'], True);
      Viewport1.CameraBookmark:=TT['camera'];
      Lathe.CalcMesh;
    finally
      FreeAndNil(TT);
    end;
  end;

begin
  if MessageDlg('Open File', 'Are you sure? Any modifications you have made will be lost!', mtConfirmation, mbYesNo, 0) <> mrYes then Exit;
  OpenDialog1.FileName:=FileName;
  if not OpenDialog1.Execute then Exit;
  NewFile;
  try
    OpenFrom(OpenDialog1.FileName);
    FileName:=OpenDialog1.FileName;
    UpdateCaption;
  except
    ShowMessage('Failed to open ' + OpenDialog1.FileName);
  end;
  FillUIFromLathe;
  UpdateLatheBox;
  mViewFloorGridClick(mViewFloorGrid);
  pbLightPos.Invalidate;
end;

procedure TMain.mFileSaveAsClick(Sender: TObject);
begin
  SaveDialog1.FileName:=FileName;
  if not SaveDialog1.Execute then Exit;
  SaveToFile(SaveDialog1.FileName);
end;

procedure TMain.mFileSaveClick(Sender: TObject);
begin
  if FileName='' then begin
    mFileSaveAs.Click;
    Exit;
  end;
  SaveToFile(FileName);
end;

procedure TMain.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  CanClose:=MessageDlg('Exit', 'Are you sure? Any modifications you have made will be lost!', mtConfirmation, mbYesNo, 0)=mrYes;
end;

procedure TMain.mHelpAboutClick(Sender: TObject);
begin
  ShowMessage('Little Lathe version 1.0' + LineEnding +
    'Copyright (C) 2019 Kostas Michalopoulos' + LineEnding + LineEnding +
    'Written in Free Pascal using Lazarus');
end;

procedure TMain.SetGridSize(AValue: Integer);
begin
  if FGridSize=AValue then Exit;
  FGridSize:=AValue;
  UpdateLatheBox;
end;

procedure TMain.NewFile;
begin
  SetLength(Lathe.Quads, 0);
  Lathe.Shape.Clear;
  FillChar(Lathe, SizeOf(Lathe), 0);
  Lathe.Slices:=8;
  Lathe.LightPos:=Vector(-16, 30, 0);
  Lathe.AmbientColor:=Vector(0.2, 0.2, 0.2);
  Lathe.ObjectColor:=Vector(0.7, 0.7, 0.7);
  Lathe.LightColor:=Vector(1.0, 1.0, 1.0);
  Lathe.SharpAngle:=0;
  Lathe.AmbientShade:=1;
  Lathe.Scale:=1;
  Lathe.CalcMesh;
  GridSize:=16;
  BgColor:=clBlack;
  Viewport1.CameraBookmark:='[0 16 64 0 0 -1 1 perspective]';
  if not mViewFloorGrid.Checked then mViewFloorGrid.Click;
  UpdateLatheBox;
  pbLightPos.Invalidate;
  FillUIFromLathe;
end;

procedure TMain.EnsureBitmapSize;
begin
  if Assigned(DBuffer) and ((pbLathe.Width <> DBuffer.Width) or (pbLathe.Height <> DBuffer.Height)) then
    FreeAndNil(DBuffer);
  if not Assigned(DBuffer) then begin
    DBuffer:=TBitmap.Create;
    DBuffer.PixelFormat:=pf24bit;
    DBuffer.Width:=pbLathe.Width;
    DBuffer.Height:=pbLathe.Height;
  end;
end;

procedure TMain.UpdateLatheBox;
begin
  {$IFDEF DARWIN}
  pbLathe.Invalidate;
  {$ELSE}
  DrawLathe;
  {$ENDIF}
end;

procedure TMain.PickLatheBox(const P: TVector2D);
begin
  LatheBoxPickResult:=lbprEmpty;
  LatheBoxPickIndex:=Lathe.ClosestPointAt(P, 4);
  if LatheBoxPickIndex <> -1 then LatheBoxPickResult:=lbprPoint;
end;

procedure TMain.DrawLathe;
var
  I, X, Y: Integer;
begin
  EnsureBitmapSize;
  with DBuffer.Canvas do begin
    Brush.Color:=clWhite;
    Pen.Color:=clWhite;
    Rectangle(0, 0, pbLathe.Width, pbLathe.Height);
    Pen.Color:=$E0E0E0;
    if GridSize > 2 then begin
      Y:=pbLathe.Height - 1;
      while Y >= 0 do begin
        Line(0, Y, pbLathe.Width, Y);
        Dec(Y, GridSize);
      end;
      X:=0;
      while X < pbLathe.Width do begin
        Line(X, 0, X, pbLathe.Height);
        Inc(X, GridSize);
      end;
      Y:=pbLathe.Height - 1;
      while Y >= 0 do begin
        X:=0;
        while X < pbLathe.Width do begin
          Pixels[X, Y]:=$404040;
          Inc(X, GridSize);
        end;
        Dec(Y, GridSize);
      end;
    end;
    Pen.Color:=clBlack;
    with Lathe.Shape do begin
      for I:=1 to High(Vertices) do begin
        Line(Round(Vertices[I - 1].X), Round(pbLathe.Height - 1 - Vertices[I - 1].Y),
             Round(Vertices[I].X), Round(pbLathe.Height - 1 - Vertices[I].Y));
      end;
      Brush.Color:=clBlack;
      for I:=0 to High(Vertices) do begin
        Rectangle(Round(Vertices[I].X - 1), Round(pbLathe.Height - 1 - Vertices[I].Y - 1),
                  Round(Vertices[I].X + 2), Round(pbLathe.Height - 1 - Vertices[I].Y + 2));
      end;
    end;
  end;
  pbLathe.Canvas.Draw(0, 0, DBuffer);
end;

procedure TMain.Render;
var
  I, J: Integer;
begin
  glEnable(GL_CULL_FACE);
  glBegin(GL_QUADS);
  with Lathe do begin
    for I:=0 to High(Quads) do begin
      for J:=0 to 3 do begin
        glColor3d(Quads[I][J].C.x, Quads[I][J].C.y, Quads[I][J].C.z);
        glVertex3d(Quads[I][J].P.x, Quads[I][J].P.y, Quads[I][J].P.z);
      end;
    end;
  end;
  glEnd();
  glDisable(GL_CULL_FACE);
end;

procedure TMain.DeleteVertex(Index: Integer);
begin
  if (Index < 0) or (Index > High(Lathe.Shape.Vertices)) then Exit;
  Lathe.Shape.Remove(Index);
  Lathe.CalcMesh;
  UpdateLatheBox;
end;

procedure TMain.SplitVertex(Index: Integer);
var
  P: TVector2D;
  I: Integer;
begin
  if (Index < 0) or (Index >= High(Lathe.Shape.Vertices)) then Exit;
  Lathe.Shape.Add(Lathe.Shape.Vertices[High(Lathe.Shape.Vertices)]);
  P:=MidpointBetween(Lathe.Shape.Vertices[Index], Lathe.Shape.Vertices[Index + 1]).Snapped(1);
  for I:=High(Lathe.Shape.Vertices) downto Index + 2 do
    Lathe.Shape.Vertices[I]:=Lathe.Shape.Vertices[I - 1];
  Lathe.Shape.Vertices[Index + 1]:=P;
  Lathe.CalcMesh;
  UpdateLatheBox;
end;

procedure TMain.FillUIFromLathe;
begin
  cbStyle.ItemIndex:=Ord(Lathe.Smooth);
  cbSlices.Text:=IntToStr(Lathe.Slices) + ' Slices';
  tbScale.Position:=Round(1000*Lathe.Scale);
  tbSharpness.Position:=Round(1000*Lathe.SharpAngle);
  cbObjectColor.ButtonColor:=VectorToColor(Lathe.ObjectColor);
  cbAmbientColor.ButtonColor:=VectorToColor(Lathe.AmbientColor);
  cbLightColor.ButtonColor:=VectorToColor(Lathe.LightColor);
  tbAmbientShade.Position:=Round(1000*Lathe.AmbientShade);
end;

procedure TMain.UpdateCaption;
begin
  if FileName='' then Caption:='Little Lathe' else Caption:=ExtractFileName(FileName) + ' - Little Lathe';
end;

procedure TMain.SaveToFile(AFileName: string);
  procedure SaveTo(AFileName: string);
  var
    TT: TTreeText;
    I: Integer;
  begin
    TT:=TTreeText.Create(Self);
    try
      TT['points']:=IntToStr(Length(Lathe.Shape.Vertices));
      for I:=0 to High(Lathe.Shape.Vertices) do begin
        TT['p' + IntToStr(I) + 'x']:=IntToStr(Round(Lathe.Shape.Vertices[I].X));
        TT['p' + IntToStr(I) + 'y']:=IntToStr(Round(Lathe.Shape.Vertices[I].Y));
      end;
      TT['slices']:=IntToStr(Lathe.Slices);
      TT['light.x']:=FloatToString(Lathe.LightPos.x);
      TT['light.y']:=FloatToString(Lathe.LightPos.y);
      TT['ambient.r']:=FloatToString(Lathe.AmbientColor.x);
      TT['ambient.g']:=FloatToString(Lathe.AmbientColor.y);
      TT['ambient.b']:=FloatToString(Lathe.AmbientColor.z);
      TT['object.r']:=FloatToString(Lathe.ObjectColor.x);
      TT['object.g']:=FloatToString(Lathe.ObjectColor.y);
      TT['object.b']:=FloatToString(Lathe.ObjectColor.z);
      TT['light.r']:=FloatToString(Lathe.LightColor.x);
      TT['light.g']:=FloatToString(Lathe.LightColor.y);
      TT['light.b']:=FloatToString(Lathe.LightColor.z);
      TT['sharp']:=FloatToString(Lathe.SharpAngle);
      TT['ambient.shade']:=FloatToString(Lathe.AmbientShade);
      TT['scale']:=FloatToString(Lathe.Scale);
      TT['smooth']:=BoolToStr(Lathe.Smooth);
      TT['bgcolor.r']:=FloatToString(ColorToVector(BgColor).x);
      TT['bgcolor.g']:=FloatToString(ColorToVector(BgColor).y);
      TT['bgcolor.b']:=FloatToString(ColorToVector(BgColor).z);
      TT['floorgrid']:=BoolToStr(mViewFloorGrid.Checked);
      TT['camera']:=Viewport1.CameraBookmark;
      TT.SaveToFile(AFileName);
    finally
      FreeAndNil(TT);
    end;
  end;

begin
  try
    SaveTo(AFileName);
    FileName:=AFileName;
    UpdateCaption;
  except
    ShowMessage('Failed to save ' + AFileName);
  end;
end;

end.

